
module type Type =
sig
	type t = 
	| X			 (* the x variable *)
	| R of float (* a real number  *)
	| Add of t * t
	| Mul of t * t
	| Sin of t
	| Cos of t
	| Tan of t
	| Log of t
	| Exp of t
	| Power of t * float
	val derive : t -> t
end

module Data
:
Type
=
struct

	type t = 
	| X			 (* the x variable *)
	| R of float (* a real number  *)
	| Add of t * t
	| Mul of t * t
	| Sin of t
	| Cos of t
	| Tan of t
	| Log of t
	| Exp of t
	| Power of t * float

let factor = function
  | R(1.),u -> u
  | u,v -> Mul(u,v)

let product = function
  | R(a),Mul(R(b),u) -> factor(R(a*.b),u)
  | Mul(R(a),u),R(b) -> factor(R(a*.b),u)
  | Mul(R(a),u),Mul(R(b),v) -> factor(R(a*.b),Mul(u,v))
  | u,Mul(R(k),v) -> Mul(R(k),Mul(u,v))
  | Mul(R(k),u),v -> Mul(R(k),Mul(u,v))
  | u,v -> factor(u,v)	

let rec derive = function
	| X -> R(1.)
	| R(k) -> R(0.)
	| Add(u,R(k)) -> derive(u)
	| Add(u,v)	  -> Add(derive(u),derive(v))
	| Mul(R(k),X) -> R(k)
	| Mul(R(k),u) -> product(R(k),derive(u))
	| Mul(u,v)	  -> Add(product(derive(u),v),product(u,derive(v)))
	| Sin(u) -> product(derive(u),Cos(u))
	| Cos(u) -> product(R(-1.),product(derive(u),Sin(u)))
	| Tan(u) -> product(derive(u),Power(Cos(u),-2.))
	| Log(u) -> product(derive(u),Power(u,-1.))
	| Exp(u) -> product(derive(u),Exp(u))
	| Power(u,2.) -> product(R(2.),product(derive(u),u))
	| Power(u,a)  -> product(R(a),product(derive(u),Power(u,a-.1.)))
end


(* using a recursor *)

type t = 
	| X			 (* the x variable *)
	| R of float (* a real number  *)
	| Add of t * t
	| Mul of t * t
	| Sin of t
	| Cos of t
	| Tan of t
	| Log of t
	| Exp of t
	| Power of t * float

(* recursion for derive *)
let rec recu c = function
	| X -> c#x
	| R k -> c#r
	| Add (u,v) -> c#add (recu c u) (recu c v)
	| Mul (u,v) -> c#mul u (recu c u) v (recu c v)
	| Sin u -> c#sin u (recu c u)
	| Cos u -> c#cos u (recu c u)
	| Tan u -> c#tan u (recu c u)
	| Log u -> c#log u (recu c u)
	| Exp u -> c#exp u (recu c u)
	| Power (u,2.) -> c#power_2 u (recu c u)
	| Power (u,a)  -> c#power_a u a (recu c u)

(* laws for derive *)
let derive =  
	recu
		object 
			method x = R 1.
			method r = R 0.
			method add du dv = Add (du,dv)
			method mul u du v dv = Add (Mul(du,v),Mul (u,dv))
			method sin u du = Mul (du,Cos(u))
			method cos u du = Mul (R(-1.),Mul (du,Sin(u)))
			method tan u du = Mul (du,Power(Cos(u),-2.))
			method log u du = Mul (du,Power(u,-1.))
			method exp u du = Mul (du,Exp u)
			method power_2 u du = Mul(R(2.),Mul(du,u))
			method power_a u a du = Mul(R(a),Mul(du,Power(u,a-.1.)))
		end
		